#include <string.h>
#include "config-consoles.h"
#include "client-message.h"
#include "spec-console.h"
#include "error.h"

typedef struct _ConfigConsoleDialog {
    Conc *conc;
    gchar *name; /* NULL if adding new console */
    GladeXML *xml; /* NULL if removing console */
} ConfigConsoleDialog;

/*
 * Glade widget names:
 */
#define CONFIG_CONSOLE_DIALOG		"config console dialog"
#define	CONFIG_CONSOLE_DIALOG_NAME	"name"
#define	CONFIG_CONSOLE_DIALOG_SERVER	"server"
#define	CONFIG_CONSOLE_DIALOG_DEVICE	"device"
#define	CONFIG_CONSOLE_DIALOG_SPEED	"speed"
#define	CONFIG_CONSOLE_DIALOG_BITS	"bits"


/*
 * Glade handler names:
 */
#define CONFIG_USER_DIALOG_RESPONSE	"response"

static void
config_consoles_done(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    ConfigConsoleDialog *ccd = data;
    GError *e = message_arg_error_check(arg);
    gchar *failed_title = _("Failed to update console");
    gchar *failed_text = _("Could not modify console (%s)");

    if (ccd->xml == NULL) {
	failed_title = _("Failed to remove console");
	failed_text = _("Could not remove console (%s)");
    }
    if (ccd->name == NULL) {
	failed_title = _("Failed to add console");
	failed_text = _("Could not add console (%s)");
    }
    if (e != NULL) {
	warning_dialog(ccd->conc, failed_title, failed_text, e->message);
	g_clear_error(&e);
	if (ccd->xml != NULL) {
	    return;
	}
    }
    if (ccd->xml) {
	gtk_widget_destroy(glade_xml_get_widget(ccd->xml, CONFIG_CONSOLE_DIALOG));
    }
    if (ccd->name) {
	g_free(ccd->name);
    }
    g_free(ccd);
}

static void
config_consoles_dialog_response(GtkDialog *dialog,
	gint arg1,
	gpointer user_data) {
    ConfigConsoleDialog *ccd = user_data;

    if (arg1 == GTK_RESPONSE_OK) {
	gchar *command;
	SpecConsole *sc;
	MessageArg *a;
	GError *e = NULL;
	const gchar *name;
	const gchar *device;
	gchar *server;
	gchar *speed;
	gchar *bits;

	name = gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(ccd->xml,
			CONFIG_CONSOLE_DIALOG_NAME)));
	if (name == NULL || strlen(name) == 0) {
	    warning_dialog(ccd->conc, _("Cannot add console"),
		    _("No name was given"));
	    return;
	}
	device = gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(ccd->xml,
			CONFIG_CONSOLE_DIALOG_DEVICE)));
	if (device == NULL || strlen(device) == 0) {
	    warning_dialog(ccd->conc, _("Cannot add console"),
		    _("No device was given"));
	    return;
	}

	server = get_combobox_text(ccd->xml, CONFIG_CONSOLE_DIALOG_SERVER,
		GTK_TREE_MODEL(ccd->conc->all_servers), SERVERS_COLUMN_NAME);
	g_assert(server);
	speed = get_combobox_text(ccd->xml, CONFIG_CONSOLE_DIALOG_SPEED,
		NULL, 0);
	g_assert(speed);
	bits = get_combobox_text(ccd->xml, CONFIG_CONSOLE_DIALOG_BITS,
		NULL, 0);
	g_assert(bits);
	
	sc = spec_console_new(name, server, device, g_strtod(speed, NULL), bits);
	g_free(server);
	g_free(speed);
	g_free(bits);
	a = spec_console_message_arg("spec", sc);
	spec_console_free(sc);

	if (ccd->name == NULL) {
	    command = CLIENT_MESSAGE_CONFIG_CONSOLE_ADD;
	} else {
	    MessageArg *spec = a;

	    command = CLIENT_MESSAGE_CONFIG_CONSOLE_EDIT;
	    a = message_arg_list("", NULL);
	    a->list = g_list_prepend(a->list,
		    message_arg_string("name", ccd->name));
	    a->list = g_list_prepend(a->list, spec);
	}
	if (message_send(ccd->conc->msg_q, command, a, config_consoles_done,
		    ccd, &e) == FALSE) {
	    error_dialog(ccd->conc, _("Internal error"),
		    _("Failed to send message (%s)"), e->message);
	    g_clear_error(&e);
	}
	message_arg_free(a);
	return;
    }
    gtk_widget_destroy(glade_xml_get_widget(ccd->xml, CONFIG_CONSOLE_DIALOG));
    if (ccd->name) {
	g_free(ccd->name);
    }
    g_free(ccd);
}

static void
config_consoles_dialog(ConfigConsoleDialog *ccd) {
    GtkCellRenderer *text_rend;

    ccd->xml = glade_xml_new(ccd->conc->glade_file, CONFIG_CONSOLE_DIALOG,
	    NULL);
    if (ccd->name != NULL) {
	gchar *title = g_strdup_printf(_("Edit Console: %s"), ccd->name);

	g_object_set(glade_xml_get_widget(ccd->xml, CONFIG_CONSOLE_DIALOG),
		    "title", title, FALSE);
    }
    g_object_set(glade_xml_get_widget(ccd->xml, CONFIG_CONSOLE_DIALOG_SERVER),
	    "model", ccd->conc->all_servers, NULL);
    text_rend = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(glade_xml_get_widget(ccd->xml,
		    CONFIG_CONSOLE_DIALOG_SERVER)), text_rend, TRUE);
    gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(glade_xml_get_widget(ccd->xml,
		    CONFIG_CONSOLE_DIALOG_SERVER)), text_rend, "text",
	    SERVERS_COLUMN_NAME);
    gtk_combo_box_set_active(GTK_COMBO_BOX(glade_xml_get_widget(ccd->xml,
		    CONFIG_CONSOLE_DIALOG_SERVER)), 0);
    gtk_combo_box_set_active(GTK_COMBO_BOX(glade_xml_get_widget(ccd->xml,
		    CONFIG_CONSOLE_DIALOG_SPEED)), 4);
    gtk_combo_box_set_active(GTK_COMBO_BOX(glade_xml_get_widget(ccd->xml,
		    CONFIG_CONSOLE_DIALOG_BITS)), 0);
    glade_xml_signal_connect_data(ccd->xml, CONFIG_USER_DIALOG_RESPONSE,
	    (GCallback)config_consoles_dialog_response, ccd);
}

extern void
config_consoles_add(Conc *conc) {
    ConfigConsoleDialog *ccd = g_new0(struct _ConfigConsoleDialog, 1);

    ccd->conc = conc;
    ccd->name = NULL;
    config_consoles_dialog(ccd);
}

static void
config_consoles_edit_data(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    ConfigConsoleDialog *ccd = data;
    GError *e = message_arg_error_check(arg);
    SpecConsole *sc;
    gchar *s;

    if (e != NULL) {
	warning_dialog(ccd->conc, _("Failed to edit console"),
		_("Could not edit console (%s)"), e->message);
	g_clear_error(&e);
	if (ccd->name) {
	    g_free(ccd->name);
	}
	g_free(ccd);
	return;
    }
    config_consoles_dialog(ccd);
    sc = spec_console_new_from_arg(arg);
    g_assert(sc);
    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(ccd->xml,
		    CONFIG_CONSOLE_DIALOG_NAME)), sc->name);
    set_combobox_text(ccd->xml, CONFIG_CONSOLE_DIALOG_SERVER,
	    GTK_TREE_MODEL(ccd->conc->all_servers), SERVERS_COLUMN_NAME,
	    sc->server);
    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(ccd->xml,
		    CONFIG_CONSOLE_DIALOG_DEVICE)), sc->terminal->device);
    s = g_strdup_printf("%d", sc->terminal->ispeed);
    set_combobox_text(ccd->xml, CONFIG_CONSOLE_DIALOG_SPEED, NULL, 0, s);
    g_free(s);
    s = "8N1";
    if (sc->terminal->data_bits == 7 && sc->terminal->parity == PARITY_EVEN &&
	    sc->terminal->stop_bits == 2) {
	s = "7E2";
    }
    set_combobox_text(ccd->xml, CONFIG_CONSOLE_DIALOG_BITS, NULL, 0, s);
    spec_console_free(sc);
}

extern void
config_consoles_edit(Conc *conc,
	const gchar *name) {
    ConfigConsoleDialog *ccd = g_new0(struct _ConfigConsoleDialog, 1);
    MessageArg *a;
    GError *e = NULL;

    ccd->conc = conc;
    ccd->name = g_strdup(name);
    ccd->xml = NULL;
    a = message_arg_string("name", ccd->name);
    if (message_send(ccd->conc->msg_q, CLIENT_MESSAGE_CONFIG_CONSOLE_SPEC,
		a, config_consoles_edit_data, ccd, &e) == FALSE) {
	error_dialog(ccd->conc, _("Internal error"),
		_("Failed to send message (%s)"), e->message);
	g_clear_error(&e);
    }
    message_arg_free(a);
}

extern void
config_consoles_remove(Conc *conc,
	const gchar *name) {
    ConfigConsoleDialog *ccd = g_new0(struct _ConfigConsoleDialog, 1);
    MessageArg *a;
    GError *e = NULL;

    ccd->conc = conc;
    ccd->name = g_strdup(name);
    ccd->xml = NULL;
    a = message_arg_string("name", name);
    if (message_send(ccd->conc->msg_q, CLIENT_MESSAGE_CONFIG_CONSOLE_REMOVE,
		a, config_consoles_done, ccd, &e) == FALSE) {
	error_dialog(ccd->conc, _("Internal error"),
		_("Failed to send message (%s)"), e->message);
	g_clear_error(&e);
    }
    message_arg_free(a);
}


